from bs4 import BeautifulSoup
from bs4.dammit import EntitySubstitution
from bs4.element import Comment
from bs4.element import Doctype
from bs4.element import NavigableString
from bs4.element import ProcessingInstruction
from bs4.element import Tag
from zpretty.attributes import PrettyAttributes
from zpretty.text import endswith_whitespace
from zpretty.text import lstrip_first_line
from zpretty.text import rstrip_last_line
from zpretty.text import startswith_whitespace


class OpenTagException(Exception):
    """We want this element to be closed"""

    def __init__(self, el):
        """el is a PrettyElement instance"""
        self.el = el

    def __str__(self):
        return "Known self closing tag %r is not closed" % self.el.context


def memo(f):
    """Simple memoize"""
    key = "__zpretty_memo__" + f.__name__

    def wrapped(obj):
        if not hasattr(obj, key):
            setattr(obj, key, f(obj))
        return getattr(obj, key)

    return wrapped


class PrettyElement(object):
    """A pretty element class that can render prettified html"""

    null_tag_name = "null_tag_name"

    knownself_closing_elements = [
        "area",
        "base",
        "basefont",
        "br",
        "col",
        "embed",
        "frame",
        "hr",
        "img",
        "input",
        "link",
        "meta",
        "param",
        "source",
        "track",
        "wbr",
    ]
    indent = "  "
    attribute_klass = PrettyAttributes

    first_attribute_on_new_line = False
    before_closing_multiline = ""

    self_closing_singleline_attributeless_template = "{prefix}<{tag} />"
    self_closing_singleline_attributefull_template = "{prefix}<{tag} {attributes} />"
    self_closing_multiline_template = "\n".join(
        ("{prefix}<{tag} {attributes}", "{prefix}{before_closing_multiline}/>")
    )

    start_tag_singleline_attributeless_template = "{prefix}<{tag}>"
    start_tag_singleline_attributefull_template = "{prefix}<{tag} {attributes}>"
    start_tag_multiline_template = "\n".join(
        ("{prefix}<{tag} {attributes}", "{prefix}{before_closing_multiline}>")
    )
    escaper = EntitySubstitution()
    preserve_text_whitespace_elements = ["pre"]
    skip_text_escaping_elements = ["script", "style"]

    def __init__(self, context, level=0):
        """Take something a (bs4) element and an indentation level"""
        self.context = context
        self.level = level

    def __str__(self):
        """Reuse the context method"""
        return str(self.context)

    def __repr__(self):
        """Try to make evident:

        - the element type
        - the level
        """
        if self.is_comment():
            tag = "!--"
        if self.is_text():
            tag = '""'
        else:
            tag = self.tag
        return "<pretty:{level}:{tag} />".format(tag=tag, level=self.level)

    def is_comment(self):
        """Check if this element is a comment"""
        return isinstance(self.context, Comment)

    def is_doctype(self):
        """Check if this element is a doctype"""
        return isinstance(self.context, Doctype)

    def is_text(self):
        """Check if this element is a text

        Also comments and processing instructions
        are instances of NavigableString,
        so we have to make additional checks
        """
        if not isinstance(self.context, NavigableString):
            return False
        if self.is_comment() or self.is_doctype() or self.is_processing_instruction():
            return False
        return True

    def is_tag(self):
        """Check if this element is a notmal tag"""
        return isinstance(self.context, Tag)

    def is_self_closing(self):
        """Is this element self closing?"""
        if not self.is_tag():
            raise ValueError("This is not a tag")
        # First check if element has some content.
        # If it has it cannot be self closing
        tag_name = self.tag
        if self.getchildren():
            if tag_name in self.knownself_closing_elements:
                raise OpenTagException(self)
            return False
        # Then we have some know elements that we want to be self closing
        if tag_name in self.knownself_closing_elements:
            return True
        # Also elements in the tal namespace may be prettified as self closing
        # if needed, e.g.: <tal:name replace="${here/title}" />
        if tag_name.startswith("tal:"):
            return True
        if tag_name.startswith("metal:"):
            return True
        # All the other elements will have an open an close tag
        return False

    def is_null(self):
        """We define a special tag null_tag_name to wrap text"""
        return self.context.name == self.null_tag_name

    def is_soup(self):
        """Check if this element is a BeautifulSoup instance"""
        return isinstance(self.context, BeautifulSoup)

    def is_processing_instruction(self):
        """Check if this element is a processing instruction like <?xml...>"""
        return isinstance(self.context, ProcessingInstruction)

    @memo
    def getparent(self):
        """Return the element parent as an instance of this class"""
        parent = self.context.parent
        if not parent or parent.name == BeautifulSoup.ROOT_TAG_NAME:
            return None
        return self.__class__(parent)

    @memo
    def getchildren(self):
        """Return this element children as instances of this class"""
        children = []
        next_level = self.level + 1
        for child in getattr(self.context, "children", []):
            child = self.__class__(child, next_level)
            try:
                child.is_tag() and child.is_self_closing()
            except OpenTagException:
                # Fix open tags and repeat
                nephews = reversed(tuple(child.context.children))
                for nephew in nephews:
                    child.context.insert_after(nephew)
                return self.getchildren()
            children.append(child)
        return children

    @property
    def tag(self):
        """Return the tag name"""
        return self.context.name

    @property
    def text(self):
        """Return the text contained in this element (if any)

        Convert the text characters to html entities
        """
        if not isinstance(self.context, NavigableString):
            return ""
        if self.is_comment():
            return self.context
        if (
            not self.escaper
            or self.context.parent.name in self.skip_text_escaping_elements
        ):
            return self.context.string
        return self.escaper.substitute_html(self.context.string)

    @property
    @memo
    def attributes(self):
        """Return the wrapped attributes"""
        attributes = getattr(self.context, "attrs", {})
        return self.attribute_klass(attributes, self)

    @property
    @memo
    def preserve_text_whitespace(self):
        children = self.getchildren()
        return (
            self.tag in self.preserve_text_whitespace_elements
            and len(children) == 1
            and children[0].is_text()
        )

    @memo
    def render_content(self):
        """Render a properly indented the contents of this element"""
        parts = []
        previous_part = ""

        if self.preserve_text_whitespace:
            for child in self.getchildren():
                return child.text

        for idx, child in enumerate(self.getchildren()):
            part = child()
            if child.is_text():
                part = lstrip_first_line(part)
            elif not endswith_whitespace(previous_part):
                part = lstrip_first_line(part)
            else:
                parts[-1] = rstrip_last_line(parts[-1])
            parts.append(part)
            previous_part = child()
        content = "".join(parts)

        if endswith_whitespace(content):
            content = rstrip_last_line(content)
        return content

    @property
    def prefix(self):
        return self.indent * self.level

    def render_comment(self):
        """Render a properly indented comment"""
        return f"{self.prefix}<!--{self.text}-->"

    def render_doctype(self):
        """Render a properly indented comment"""
        doctype = f"{self.prefix}{self.context.PREFIX}{self.text}{self.context.SUFFIX}"
        if isinstance(
            self.context.nextSibling, NavigableString
        ) and self.context.nextSibling.startswith("\n"):
            doctype = doctype.rstrip()
        return doctype

    def render_processing_instruction(self):
        """Render a properly indented processing instruction"""
        return f"{self.prefix}<?{self.text.rstrip('?')}?>"

    def render_soup(self):
        first_child = next(self.context.children)
        if isinstance(first_child, Doctype) and not self.context.is_xml:
            return self.render_content()
        if isinstance(first_child, ProcessingInstruction):
            return self.render_content()
        return f'<?xml version="1.0" encoding="utf-8"?>\n{self.render_content()}'

    def render_text(self):
        """Render a properly indented text

        If the text starts with spaces, strip them and add a newline.
        If the text end with spaces, strip them.
        """
        text = self.text
        lines = text.split("\n")
        if not lines:
            return ""

        prefix = self.prefix
        if len(lines) == 1:
            line = lines[0]
            if not line.strip():
                return "\n"
            if startswith_whitespace(line):
                line = f"\n{prefix}{line.lstrip()}"
            if endswith_whitespace(line):
                line = f"{line.rstrip()}\n"
            return line

        if not lines[0].strip():
            rendered_lines = ["\n"]
        elif startswith_whitespace(lines[0]):
            rendered_lines = [f"\n{prefix}{lines[0].rstrip()}\n"]
        else:
            rendered_lines = [f"{lines[0]}\n"]

        for line in lines[1:-1]:
            if not line.strip():
                rendered_lines.append("\n")
            else:
                rendered_lines.append(f"{line.rstrip()}\n")

        if lines[-1].strip():
            if lines[-1].rstrip() == lines[-1]:
                if lines[-1].lstrip() == lines[-1]:
                    rendered_lines.append(lines[-1])
                else:
                    rendered_lines.append(prefix + lines[-1].lstrip())
            else:
                rendered_lines.append("%s\n" % lines[-1].rstrip())
        else:
            rendered_lines.append("")
        text = "".join(rendered_lines)
        return text

    def _render_template(self, template):
        return template.format(
            before_closing_multiline=self.before_closing_multiline,
            attributes=self.attributes.lstrip(),
            prefix=self.prefix,
            tag=self.tag,
        )

    def render_self_closing(self):
        """Render a properly indented a self closing tag"""
        attributes_len = len(self.attributes)
        if attributes_len == 0:
            template = self.self_closing_singleline_attributeless_template
        elif attributes_len == 1:
            template = self.self_closing_singleline_attributefull_template
        else:
            template = self.self_closing_multiline_template
        return self._render_template(template)

    def render_not_self_closing(self):
        """Render a properly indented not self closing tag"""
        attributes_len = len(self.attributes)
        if attributes_len == 0:
            open_tag_template = self.start_tag_singleline_attributeless_template
        elif attributes_len == 1:
            open_tag_template = self.start_tag_singleline_attributefull_template
        else:
            open_tag_template = self.start_tag_multiline_template

        text = self.text and self.render_text() or self.render_content()

        if not self.preserve_text_whitespace and endswith_whitespace(text):
            if text[-1] != "\n":
                text = f"{rstrip_last_line(text)}\n"
            close_tag_template = "{prefix}</{tag}>"
        else:
            close_tag_template = "</{tag}>"

        open_tag = self._render_template(open_tag_template)
        close_tag = self._render_template(close_tag_template)
        return "{open_tag}{text}{close_tag}".format(
            close_tag=close_tag, open_tag=open_tag, text=text
        )

    @memo
    def __call__(self):
        """Render the element and its contents properly indented"""
        if self.is_soup():
            return self.render_soup()

        if self.is_null():
            return self.render_content()

        if self.is_comment():
            return self.render_comment()

        if self.is_tag():
            if self.is_self_closing():
                return self.render_self_closing()
            else:
                return self.render_not_self_closing()

        if self.is_processing_instruction():
            return self.render_processing_instruction()

        if self.is_doctype():
            return self.render_doctype()

        return self.render_text()



from pkg_resources import resource_filename
from tempfile import TemporaryDirectory
from unittest import TestCase
from zpretty.prettifier import ZPrettifier
from zpretty.tests.mock import MockCLIRunner
from zpretty.xml import XMLPrettifier
from zpretty.zcml import ZCMLPrettifier

import os


class TestCli(TestCase):
    """Test the cli options"""

    def test_defaults(self):
        config = MockCLIRunner().config
        self.assertEqual(config.paths, "-")
        self.assertFalse(config.inplace)
        self.assertFalse(config.xml)
        self.assertFalse(config.zcml)
        self.assertEqual(config.encoding, "utf8")
        self.assertFalse(config.check)

    def test_short_options(self):
        config = MockCLIRunner("-i", "-x", "-z").config
        self.assertTrue(all((config.inplace, config.xml, config.zcml)))

    def test_long_options(self):
        config = MockCLIRunner("--inplace", "--xml", "--zcml").config
        self.assertTrue(all((config.inplace, config.xml, config.zcml)))

    def test_file(self):
        html = resource_filename("zpretty.tests", "original/sample_html.html")
        xml = resource_filename("zpretty.tests", "original/sample_xml.xml")
        config = MockCLIRunner(html, xml).config
        self.assertEqual(config.paths, [html, xml])

    def test_stdin(self):
        clirunner = MockCLIRunner()
        self.assertListEqual(clirunner.good_paths, ["-"])

    def test_broken_file_path(self):
        with TemporaryDirectory() as tmpdir:
            bad_path = os.path.join(tmpdir, "bad path")
            good_path = os.path.join(tmpdir, "good path")

            with open(good_path, "w") as f:
                f.write("I do exist")

            clirunner = MockCLIRunner(bad_path, good_path)
            self.assertListEqual(clirunner.good_paths, [good_path])

    def test_choose_prettifier(self):
        """Check the for the given options and file the best choice is made"""
        clirunner = MockCLIRunner("--xml", "--zcml")
        self.assertEqual(clirunner.choose_prettifier(""), ZCMLPrettifier)

        clirunner = MockCLIRunner("--xml")
        self.assertEqual(clirunner.choose_prettifier(""), XMLPrettifier)

        # Without options the extension rules
        clirunner = MockCLIRunner()
        self.assertEqual(clirunner.choose_prettifier("a.zcml"), ZCMLPrettifier)
        self.assertEqual(clirunner.choose_prettifier("a.xml"), XMLPrettifier)
        # The default one is returned if the extension is not recognized
        self.assertEqual(clirunner.choose_prettifier("a.txt"), ZPrettifier)

    def test_check(self):
        config = MockCLIRunner("--check").config
        self.assertTrue(config.check)

    def test_run_check(self):
        # XXX increase coverage by improving the mock
        from unittest import mock

        clirunner = MockCLIRunner("--check", "zpretty/tests/original/sample_xml.xml")
        with mock.patch("builtins.exit", return_value=None) as mocked:
            clirunner.run()
            mocked.assert_not_called()

        clirunner = MockCLIRunner("--check", "broken/broken.xml")

        with mock.patch("builtins.exit", return_value=None) as mocked:
            clirunner.run()
            mocked.assert_called_once_with(1)

    def test_good_paths(self):
        """Test the good_paths property"""
        clirunner = MockCLIRunner()
        self.assertListEqual(clirunner.good_paths, ["-"])

        clirunner = MockCLIRunner("zpretty/tests/original/sample_xml.xml")
        self.assertListEqual(
            clirunner.good_paths, ["zpretty/tests/original/sample_xml.xml"]
        )

        clirunner = MockCLIRunner("broken/broken.xml")
        self.assertListEqual(clirunner.good_paths, [])

        clirunner = MockCLIRunner(
            "broken/broken.xml", "zpretty/tests/original/sample_xml.xml"
        )
        self.assertListEqual(
            clirunner.good_paths, ["zpretty/tests/original/sample_xml.xml"]
        )

        clirunner = MockCLIRunner("zpretty/tests", "--include", "broken")
        self.assertListEqual(
            clirunner.good_paths,
            [
                "zpretty/tests/broken/broken.xml",
            ],
        )

        clirunner = MockCLIRunner(
            "zpretty/tests", "--include", "broken", "--exclude", "broken"
        )
        self.assertListEqual(clirunner.good_paths, [])

        clirunner = MockCLIRunner(
            "zpretty/tests/include-exclude",
        )
        self.assertListEqual(
            clirunner.good_paths,
            [
                "zpretty/tests/include-exclude/foo/bar/bar.html",
                "zpretty/tests/include-exclude/foo/bar/bar.pt",
                "zpretty/tests/include-exclude/foo/bar/bar.xml",
                "zpretty/tests/include-exclude/foo/bar/bar.zcml",
                "zpretty/tests/include-exclude/foo/bar/foo.html",
                "zpretty/tests/include-exclude/foo/bar/foo.pt",
                "zpretty/tests/include-exclude/foo/bar/foo.xml",
                "zpretty/tests/include-exclude/foo/bar/foo.zcml",
            ],
        )

        clirunner = MockCLIRunner(
            "zpretty/tests/include-exclude", "--extend-exclude", "bar/foo"
        )
        self.assertListEqual(
            clirunner.good_paths,
            [
                "zpretty/tests/include-exclude/foo/bar/bar.html",
                "zpretty/tests/include-exclude/foo/bar/bar.pt",
                "zpretty/tests/include-exclude/foo/bar/bar.xml",
                "zpretty/tests/include-exclude/foo/bar/bar.zcml",
            ],
        )

        clirunner = MockCLIRunner(
            "zpretty/tests/include-exclude",
            "--include",
            "pt$",
            "--extend-exclude",
            "bar/foo",
        )
        self.assertListEqual(
            clirunner.good_paths,
            [
                "zpretty/tests/include-exclude/foo/bar/bar.pt",
            ],
        )

        clirunner = MockCLIRunner("zpretty/tests/include-exclude", "--exclude", "foo")
        self.assertListEqual(
            clirunner.good_paths,
            [
                "zpretty/tests/include-exclude/.venv/bar.html",
                "zpretty/tests/include-exclude/.venv/bar.pt",
                "zpretty/tests/include-exclude/.venv/bar.xml",
                "zpretty/tests/include-exclude/.venv/bar.zcml",
                "zpretty/tests/include-exclude/venv/bar.html",
                "zpretty/tests/include-exclude/venv/bar.pt",
                "zpretty/tests/include-exclude/venv/bar.xml",
                "zpretty/tests/include-exclude/venv/bar.zcml",
            ],
        )

        clirunner = MockCLIRunner(
            "zpretty/tests/include-exclude", "--exclude", "(foo|bar)"
        )
        self.assertListEqual(
            clirunner.good_paths,
            [],
        )

        with TemporaryDirectory() as tempdir:
            with open(os.path.join(tempdir, "foo.pt"), "w"):
                pass
            clirunner = MockCLIRunner(
                tempdir,
                "--include",
                "*",
                "--exclude",
                "+",
                "--extend-exclude",
                "[a-0]",
            )
            self.assertListEqual(
                clirunner.good_paths,
                [f"{tempdir}/foo.pt"],
            )
            self.assertListEqual(
                clirunner.errors,
                [
                    "Invalid regular expression for --exclude: '+'",
                    "Invalid regular expression for --extend-exclude: '[a-0]'",
                    "Invalid regular expression for --include: '*'",
                ],
            )



from bs4 import BeautifulSoup
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.xml import XMLElement
from zpretty.xml import XMLPrettifier


class TestZpretty(TestCase):
    """Test zpretty"""

    maxDiff = None

    def get_element(self, text, level=0):
        """Given a text return a XMLElement"""
        soup = BeautifulSoup(
            "<soup><fake_root>%s</fake_root></soup>" % text, "html.parser"
        )
        return XMLElement(soup.fake_root.next_element, level)

    def prettify(self, filename):
        """Run prettify on filename and check that the output is equal to
        the file content itself
        """
        resolved_filename = resource_filename("zpretty.tests", "original/%s" % filename)
        prettifier = XMLPrettifier(resolved_filename)
        observed = prettifier()
        expected = open(resolved_filename).read()
        self.assertListEqual(observed.splitlines(), expected.splitlines())

    def test_newline_between_attributes(self):
        """See #84"""
        element = self.get_element('<one \n foo="bar"\n\n\nbar="foo"\n/>')
        self.assertEqual(element(), '<one bar="foo"\n     foo="bar"\n/>')

    def test_zcml(self):
        self.prettify("sample_xml.xml")

    def test_sample_dtml(self):
        self.prettify("sample_dtml.dtml")

    def test_sample_txt(self):
        self.prettify("sample.txt")



from pathlib import Path
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.tests.mock import MockCLIRunner

import argparse
import sys


class TestReadme(TestCase):
    """Test zpretty"""

    maxDiff = None

    def extract_usage_from_readme(self):
        """Extract the usage from the documentation"""
        resolved_filename = Path(resource_filename("zpretty", ".")) / ".." / "README.md"

        with open(resolved_filename) as f:
            readme = f.read()
            if sys.version_info < (3, 9):  # pragma: no cover
                # Small change in the argparse output for Python >= 3.9
                readme = readme.replace("[paths ...]", "[paths [paths ...]]")
            if sys.version_info < (3, 10):  # pragma: no cover
                # Small change in the argparse output for Python >= 3.10
                readme = readme.replace("options:", "optional arguments:")
        start = readme.index("zpretty [")
        end = readme.index("```", start)
        # Take all the lines ignoring whitespaces
        return [x.strip() for x in readme[start:end].splitlines()]

    def extract_usage_from_parser(self):
        """Ask the parser for the usage and indent it"""
        parser = MockCLIRunner().parser
        # This is needed to keep the 100 lines limit
        parser.formatter_class = lambda prog: argparse.HelpFormatter(prog, width=80)
        # temporarily remove the description
        parser.description = None
        parser_help = parser.format_help()
        start = parser_help.index("zpretty [")
        parser_help = parser_help[start:]
        # Take all the lines ignoring whitespaces
        return [x.strip() for x in parser_help.splitlines()]

    def test_readme(self):
        observed = self.extract_usage_from_readme()
        expected = self.extract_usage_from_parser()
        self.assertListEqual(observed, expected)



from unittest import TestCase
from zpretty.text import endswith_whitespace
from zpretty.text import lstrip_first_line
from zpretty.text import rstrip_last_line
from zpretty.text import startswith_whitespace


class TestFunctions(TestCase):
    """Test functions used by zpretty"""

    def test_lstrip_first_line_oneline(self):
        self.assertEqual(lstrip_first_line(" a"), "a")

    def test_lstrip_first_line_twolines(self):
        self.assertEqual(lstrip_first_line(" a \n b"), ("a \n b"))

    def test_rstrip_larst_line_oneline(self):
        self.assertEqual(rstrip_last_line("a"), "a")

    def test_rstrip_larst_line_twoline(self):
        self.assertEqual(rstrip_last_line("a\n b "), ("a\n b"))

    def test_none(self):
        self.assertEqual(lstrip_first_line(None), None)
        self.assertEqual(rstrip_last_line(None), None)
        self.assertFalse(endswith_whitespace(None))
        self.assertFalse(startswith_whitespace(None))



from bs4 import BeautifulSoup
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.zcml import ZCMLAttributes
from zpretty.zcml import ZCMLElement
from zpretty.zcml import ZCMLPrettifier


class TestZpretty(TestCase):
    """Test zpretty"""

    maxDiff = None

    def get_element(self, text, level=0):
        """Given a text return a PrettyElement"""
        soup = BeautifulSoup(
            "<soup><fake_root>%s</fake_root></soup>" % text, "html.parser"
        )
        return ZCMLElement(soup.fake_root.next_element, level)

    def assertPrettifiedAttributes(self, attributes, expected, level=0):
        """Check if the attributes are properly sorted and formatted"""
        if level == 0:
            el = None
        else:
            el = self.get_element("foo", level)
        pretty_attribute = ZCMLAttributes(attributes, el)
        observed = pretty_attribute()
        self.assertEqual(observed, expected)

    def test_zcml_attributes_no_attributes(self):
        self.assertPrettifiedAttributes(ZCMLAttributes({})(), "")
        self.assertPrettifiedAttributes({}, "", level=2)

    def test_zcml_attributes_one_attributes(self):
        self.assertPrettifiedAttributes({"a": "1"}, 'a="1"')
        self.assertPrettifiedAttributes({"a": "1"}, 'a="1"', level=1)
        self.assertPrettifiedAttributes({"a": "1"}, 'a="1"', level=2)

    def test_zcml_attributes_many_attributes(self):
        self.assertPrettifiedAttributes({"a": "1", "b": "2"}, '    a="1"\n    b="2"')
        self.assertPrettifiedAttributes(
            {"a": "1", "b": "2"}, '      a="1"\n      b="2"', level=1
        )
        self.assertPrettifiedAttributes(
            {"a": "1", "b": "2"}, '        a="1"\n        b="2"', level=2
        )

    def test_zcml_self_closing_no_attributes(self):
        element = self.get_element("<zcml />")
        self.assertEqual(element(), "<zcml />")
        element = self.get_element("<zcml />", 1)
        self.assertEqual(element(), "  <zcml />")
        element = self.get_element("<zcml />", 2)
        self.assertEqual(element(), "    <zcml />")

    def test_zcml_self_closing_one_attributes(self):
        element = self.get_element('<zcml \n foo="bar"/>')
        self.assertEqual(element(), '<zcml foo="bar" />')
        element = self.get_element('<zcml \n foo="bar"/>', 1)
        self.assertEqual(element(), '  <zcml foo="bar" />')
        element = self.get_element('<zcml \n foo="bar"/>', 2)
        self.assertEqual(element(), '    <zcml foo="bar" />')

    def test_zcml_self_closing_many_attributes(self):
        element = self.get_element('<zcml \n foo="bar" bar="foo"/>')
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "<zcml",
                    '    bar="foo"',
                    '    foo="bar"',
                    "    />",
                )
            ),
        )
        element = self.get_element('<zcml \n foo="bar" bar="foo"/>', 1)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "  <zcml",
                    '      bar="foo"',
                    '      foo="bar"',
                    "      />",
                )
            ),
        )
        element = self.get_element('<zcml \n foo="bar" bar="foo"/>', 2)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "    <zcml",
                    '        bar="foo"',
                    '        foo="bar"',
                    "        />",
                )
            ),
        )

    def test_for_attribute_single_attribute(self):
        element = self.get_element('<zcml \n for="foo"/>')
        self.assertEqual(element(), '<zcml for="foo" />')
        element = self.get_element('<zcml \n for="foo"/>', 1)
        self.assertEqual(element(), '  <zcml for="foo" />')
        element = self.get_element('<zcml \n for="foo"/>', 2)
        self.assertEqual(element(), '    <zcml for="foo" />')

        element = self.get_element('<zcml \n for="foo bar"/>')
        self.assertEqual(
            element(),
            "\n".join(
                (
                    '<zcml for="foo',
                    '           bar" />',
                )
            ),
        )

        element = self.get_element('<zcml \n for="foo bar"/>', 1)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    '  <zcml for="foo',
                    '             bar" />',
                )
            ),
        )

        element = self.get_element('<zcml \n for="foo bar"/>', 2)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    '    <zcml for="foo',
                    '               bar" />',
                )
            ),
        )

    def test_for_attribute_multiple_attribute(self):
        element = self.get_element('<zcml \n for="foo" handler="bar" />')
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "<zcml",
                    '    for="foo"',
                    '    handler="bar"',
                    "    />",
                )
            ),
        )
        element = self.get_element('<zcml \n for="foo" handler="bar" />', 1)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "  <zcml",
                    '      for="foo"',
                    '      handler="bar"',
                    "      />",
                )
            ),
        )
        element = self.get_element('<zcml \n for="foo" handler="bar" />', 2)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "    <zcml",
                    '        for="foo"',
                    '        handler="bar"',
                    "        />",
                )
            ),
        )

        element = self.get_element('<zcml \n for="foo bar" handler="bar" />')
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "<zcml",
                    '    for="foo',
                    '         bar"',
                    '    handler="bar"',
                    "    />",
                )
            ),
        )
        element = self.get_element('<zcml \n for="foo bar" handler="bar" />', 1)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "  <zcml",
                    '      for="foo',
                    '           bar"',
                    '      handler="bar"',
                    "      />",
                )
            ),
        )
        element = self.get_element('<zcml \n for="foo bar" handler="bar" />', 2)
        self.assertEqual(
            element(),
            "\n".join(
                (
                    "    <zcml",
                    '        for="foo',
                    '             bar"',
                    '        handler="bar"',
                    "        />",
                )
            ),
        )

    def prettify(self, filename):
        """Run prettify on filename and check that the output is equal to
        the file content itself
        """
        resolved_filename = resource_filename("zpretty.tests", "original/%s" % filename)
        prettifier = ZCMLPrettifier(resolved_filename)
        observed = prettifier()
        expected = open(resolved_filename).read()
        self.assertListEqual(observed.splitlines(), expected.splitlines())

    def test_zcml(self):
        self.prettify("sample.zcml")



from bs4 import BeautifulSoup
from bs4.element import NavigableString
from logging import getLogger
from zpretty.attributes import PrettyAttributes
from zpretty.elements import PrettyElement
from zpretty.prettifier import ZPrettifier


logger = getLogger(__name__)


class AnyIn(object):
    def __contains__(self, item):
        return True


class XMLAttributes(PrettyAttributes):
    """Customized attribute formatter for zcml"""

    _boolean_attributes_are_allowed = False
    _known_boolean_attributes = ()
    _multiline_attributes = ()
    _tal_multiline_attributes = ()
    _xml_attribute_order = ()
    _tal_attribute_order = ()

    def sort_attributes(self, name):
        """Sort ZCML attributes in a consistent way"""
        if name in self._xml_attribute_order:
            return (100 + self._xml_attribute_order.index(name), name)
        return super(XMLAttributes, self).sort_attributes(name)


class XMLElement(PrettyElement):
    attribute_klass = XMLAttributes
    preserve_text_whitespace_elements = AnyIn()

    def is_self_closing(self):
        """Is this element self closing?"""
        if not self.is_tag():
            raise ValueError("This is not a tag")
        # Just check if the element has some content.
        return not self.getchildren()

    @property
    def tag(self):
        """Return the tag name"""
        prefix = getattr(self.context, "prefix", "")
        if not prefix:
            return self.context.name
        return f"{prefix}:{self.context.name}"

    @property
    def text(self):
        """Return the text contained in this element (if any)

        Convert the text characters to html entities
        """
        if not isinstance(self.context, NavigableString):
            return ""
        if self.is_comment():
            return self.context
        return self.escaper.substitute_xml(self.context.string)


class XMLPrettifier(ZPrettifier):
    """Prettify according to the ZCML style guide"""

    parser = "xml"
    pretty_element = XMLElement

    def get_soup(self, text):
        """Tries to get the soup from the given test

        If the text is not some xml like think a dummy element will be used to wrap it.
        """
        original_soup = BeautifulSoup(text, self.parser)
        if original_soup.is_xml:
            return original_soup

        markup = "<{null}>{text}</{null}>".format(
            null=self.pretty_element.null_tag_name, text=text
        )
        wrapped_soup = BeautifulSoup(markup, self.parser)
        return getattr(wrapped_soup, self.pretty_element.null_tag_name)



from bs4 import BeautifulSoup
from unittest import TestCase
from zpretty.elements import PrettyElement


class TestPrettyElements(TestCase):
    """Test basic funtionalities of the PrettyElement class"""

    def get_element(self, text, level=0):
        """Given a text return a PrettyElement"""
        soup = BeautifulSoup(
            "<soup><fake_root>%s</fake_root></soup>" % text, "html.parser"
        )
        return PrettyElement(soup.fake_root.next_element, level)

    def test_comment(self):
        el = self.get_element("<!--a-->")
        self.assertFalse(el.is_processing_instruction())
        self.assertFalse(el.is_doctype())
        self.assertFalse(el.is_tag())
        self.assertFalse(el.is_text())
        self.assertTrue(el.is_comment())
        self.assertEqual(el.text, "a")
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getchildren(), [])
        self.assertEqual(el.attributes(), "")
        self.assertEqual(el(), "<!--a-->")

    def test_text_element(self):
        el = self.get_element("text<ignored_children>")
        self.assertFalse(el.is_comment())
        self.assertFalse(el.is_doctype())
        self.assertFalse(el.is_processing_instruction())
        self.assertFalse(el.is_tag())
        self.assertTrue(el.is_text())
        self.assertEqual(el.text, "text")
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getchildren(), [])
        self.assertEqual(el.attributes(), "")
        self.assertEqual(el(), "text")

    def test_empty_element(self):
        el = self.get_element("<root />")
        self.assertFalse(el.is_comment())
        self.assertFalse(el.is_doctype())
        self.assertFalse(el.is_processing_instruction())
        self.assertFalse(el.is_text())
        self.assertTrue(el.is_tag())
        self.assertEqual(el.text, "")
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getchildren(), [])
        self.assertEqual(el.attributes(), "")

    def test_empty_element_attributes(self):
        el = self.get_element('<root class="b" />')
        self.assertFalse(el.is_comment())
        self.assertFalse(el.is_doctype())
        self.assertFalse(el.is_processing_instruction())
        self.assertFalse(el.is_text())
        self.assertTrue(el.is_tag())
        self.assertEqual(el.text, "")
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getchildren(), [])
        self.assertEqual(el.attributes(), 'class="b"')
        self.assertEqual(el(), '<root class="b"></root>')

    def test_processing_instruction(self):
        el = self.get_element('<?xml version="1.0" encoding="utf-8">')
        self.assertFalse(el.is_tag())
        self.assertFalse(el.is_text())
        self.assertFalse(el.is_doctype())
        self.assertTrue(el.is_processing_instruction())
        self.assertEqual(el.text, 'xml version="1.0" encoding="utf-8"')
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getchildren(), [])
        self.assertEqual(el.attributes(), "")

    def test_doctype(self):
        el = self.get_element("<!DOCTYPE html>")
        self.assertFalse(el.is_tag())
        self.assertFalse(el.is_text())
        self.assertFalse(el.is_processing_instruction())
        self.assertTrue(el.is_doctype())
        self.assertEqual(el.text, "html")
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getchildren(), [])
        self.assertEqual(el.attributes(), "")
        el = self.get_element(
            '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" '
            '"http://www.w3.org/TR/html4/strict.dtd">'
        )
        self.assertEqual(
            el.text,
            'HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" '
            '"http://www.w3.org/TR/html4/strict.dtd"',
        )

    def test_render_text(self):
        el = self.get_element(" a")
        self.assertEqual(el.render_text(), "\na")
        el = self.get_element(" ")
        self.assertEqual(el.render_text(), "\n")
        el = self.get_element("\n")
        self.assertEqual(el.render_text(), "\n")

    def test_get_parent(self):
        el = self.get_element(" a")
        self.assertEqual(el.getparent().tag, "fake_root")
        self.assertEqual(el.getparent().getparent().tag, "soup")
        self.assertIsNone(el.getparent().getparent().getparent())



from logging import getLogger


try:
    from html import escape
except ImportError:  # pragma: no cover
    # Python < 3.8
    from cgi import escape


logger = getLogger(__name__)


class PrettyAttributes(object):
    """Render attributes in a pretty way.

    - one per line
    - sorted semantically and alphabetically
    - with properly indented and escaped values
    """

    _attribute_template = "{name}={quote}{value}{quote}"
    _boolean_attributes_are_allowed = True

    _known_boolean_attributes = (
        "allowfullscreen",
        "allowpaymentrequest",
        "async",
        "autofocus",
        "autoplay",
        "checked",
        "controls",
        "default",
        "disabled",
        "formnovalidate",
        "hidden",
        "ismap",
        "itemscope",
        "loop",
        "multiple",
        "muted",
        "nomodule",
        "novalidate",
        "open",
        "playsinline",
        "readonly",
        "required",
        "reversed",
        "selected",
        "truespeed",
    )
    _multiline_prefix = "  "
    _multiline_attributes = ()
    _tal_multiline_attributes = (
        "attributes",
        "define",
        "tal:attributes",
        "tal:define",
    )

    _tal_attribute_order = (
        "tal:define",
        "tal:switch",
        "tal:condition",
        "tal:repeat",
        "tal:case",
        "tal:content",
        "tal:replace",
        "tal:omit-tag",
        "tal:attributes",
        "tal:on-error",
    )

    _i18n_attributes = (
        "i18n:translate",
        "i18n:domain",
        "i18n:context",
        "i18n:source",
        "i18n:target",
        "i18n:name",
        "i18n:attributes",
        "i18n:data",
        "i18n:comment",
        "i18n:ignore",
        "i18n:ignore-attributes",
    )

    def __init__(self, attributes, element=None):
        """attributes is a dict like object"""
        self.attributes = attributes
        self.element = element

    def __len__(self):
        return len(self.attributes)

    @property
    def prefix(self):
        """Return the prefix for the attributes

        The returned value will be a number of spaces equal to the tag name length + 2,
        e.g., in this case it will be 8
        (6 for foobar + 2 (the leading < and the space after foobar)):
        <foobar foo="1"
        ________bar="2"
        />
        """
        if not self.element:
            return "  "
        return " " * (len(self.element.tag or "") + 2)

    def sort_attributes(self, name):
        """This sorts the attribute trying to group them semantically

        Starting from the top:

        1. xml namespaces
        2. class, id
        3. attributes not belonging in to other categories (default)
        4. data- attributes
        5. tal attributes
        6. i18n attributes
        """
        if name.startswith("xmlns"):
            return (0, name)
        if name in ("class", "id"):
            return (100, name)
        if name.startswith("data"):
            return (300, name)
        if "tal:" + name in self._tal_attribute_order:
            tal_index = self._tal_attribute_order.index("tal:" + name)
            return (400 + tal_index, name)
        if name in self._tal_attribute_order:
            tal_index = self._tal_attribute_order.index(name)
            return (400 + tal_index, name)
        if name in self._i18n_attributes:
            return (900, name)
        return (200, name)

    def format_multiline(self, name, value):
        """"""
        value_lines = filter(None, value.split())
        line_joiner = "\n" + (" " * (len(name) + 2))
        return line_joiner.join(value_lines)

    def format_tal_multiline(self, value):
        """There are some tal specific attributes that contain ; separated
        statements.
        They are used to define variables or set other attributes.
        You can define many variables by adding statements separated by ';'.
        If the statement contains a ';', it will be escaped as ';;'.

        It is convenient to always have those attribute in the form:

        tal:define="
          var1 statement1;
          var2 statement2;
          ...
        "
        """
        # temp skip ';;' the escape sequence to enter a ';' in a statement
        statements = value.replace(";;", "<>").split(";")

        # We always want an empty line first...
        lines = [""]
        try:
            line_prefix = self.element.prefix + self.prefix + self._multiline_prefix
        except AttributeError:
            line_prefix = self._multiline_prefix

        for statement in statements:
            statement = statement.strip()
            if statement:
                if not statement.endswith(";"):
                    statement += ";"
                lines.append(line_prefix + statement)
        # ... and at the end dedent
        lines.append(line_prefix[:-2])

        new_value = "\n".join(lines)
        # restore ';;'
        return new_value.replace("<>", ";;")

    def is_tal_attribute(self, name):
        """Check if the attribute is a tal attribute"""
        if name.startswith("tal:"):
            return True
        try:
            if not self.element.context.name.startswith("tal:"):
                return False
        except AttributeError:
            return False
        if f"tal:{name}" in self._tal_attribute_order:
            return True

    def maybe_escape(self, name, value):
        """Escape the value if needed"""
        if self.is_tal_attribute(name):
            # Never escape what we have in tal attributes
            return value

        return escape(value, quote=False)

    def can_be_valueless(self, name):
        """Check if the attribute name can be without a value"""
        if not self._boolean_attributes_are_allowed:
            return False
        if name.startswith("data-"):
            return True
        if name in self._known_boolean_attributes:
            return True
        return False

    def lines(self):
        """Take the attributes, sort them and prettify their values"""
        attributes = self.attributes
        sorted_names = sorted(attributes, key=self.sort_attributes)
        lines = []
        for name in sorted_names:
            value = attributes[name]
            if isinstance(value, list):
                # Happens, e.g., for the class attribute
                value = " ".join(value)
            if name in self._multiline_attributes:
                value = self.format_multiline(name, value)
            elif name in self._tal_multiline_attributes:
                value = self.format_tal_multiline(value)
            if not value and self.can_be_valueless(name):
                line = name
            else:
                if '"' in value:
                    quote = "'"
                else:
                    quote = '"'

                line = self._attribute_template.format(
                    name=name, quote=quote, value=self.maybe_escape(name, value)
                )
            lines.append(line)
        return lines

    def lstrip(self):
        """This returns the attributes with the left spaces removed"""
        return self().lstrip()

    def __call__(self):
        """Render the attributes as text

        Render and an empty string if no attributes
        If we have one attribute we do not indent it
        If we have many we indent them
        """
        if len(self) == 0:
            return ""
        if len(self) == 1:
            for line in self.lines():
                return line
        if self.element:
            prefix = self.element.prefix + self.prefix
        else:
            prefix = ""
        return prefix + f"\n{prefix}".join(self.lines())



from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.prettifier import ZPrettifier


class TestZpretty(TestCase):
    """Test zpretty"""

    maxDiff = None

    def assertPrettified(self, original, expected, encoding="utf8"):
        """Check if the original html has been prettified as expected"""
        if isinstance(expected, tuple):
            expected = "\n".join(expected)
        prettifier = ZPrettifier(text=original, encoding=encoding)
        self.assertFalse(prettifier.check())
        observed = prettifier()
        self.assertEqual(observed, expected)

    def prettify(self, filename):
        """Run prettify on filename and check that the output is equal to
        the file content itself
        """
        resolved_filename = resource_filename("zpretty.tests", "original/%s" % filename)
        prettifier = ZPrettifier(resolved_filename)
        self.assertTrue(prettifier.check())
        observed = prettifier()
        expected = open(resolved_filename).read()
        self.assertListEqual(observed.splitlines(), expected.splitlines())

    def test_format_self_closing_tag(self):
        self.assertPrettified("<tal:test />", "<tal:test />\n")

    def test_nesting_no_text(self):
        # no attributes
        self.assertPrettified(
            "<root><tal:test /></root>", "<root><tal:test /></root>\n"
        )

        self.assertPrettified(
            "<root><tal:test /> </root>", "<root><tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> <tal:test /> </root>", "<root>\n  <tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> <tal:test /></root>", "<root>\n  <tal:test /></root>\n"
        )
        self.assertPrettified(
            "<root><tal:test><child /></tal:test></root>",
            "<root><tal:test><child></child></tal:test></root>\n",
        )

    def test_nesting_with_text(self):
        # no attributes
        self.assertPrettified(
            "<root> a<tal:test /></root>", "<root>\n  a<tal:test /></root>\n"
        )

        self.assertPrettified(
            "<root>a <tal:test /> </root>", "<root>a\n  <tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> a <tal:test /> </root>", "<root>\n  a\n  <tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> a <div /> </root>", "<root>\n  a\n  <div></div>\n</root>\n"
        )
        self.assertPrettified(
            "<root> a <div /> <div /> </root>",
            "<root>\n  a\n  <div></div>\n  <div></div>\n</root>\n",
        )
        self.assertPrettified(
            "<root> a <div />\n <div /> </root>",
            "<root>\n  a\n  <div></div>\n  <div></div>\n</root>\n",
        )
        self.assertPrettified(
            "<root> a <div />\na <div /> </root>",
            "<root>\n  a\n  <div></div>\na\n  <div></div>\n</root>\n",
        )
        self.assertPrettified(
            "<root> <div>a</div> </root>", "<root>\n  <div>a</div>\n</root>\n"
        )
        self.assertPrettified(
            "<root> <div>a\n  </div> </root>", "<root>\n  <div>a\n  </div>\n</root>\n"
        )
        self.assertPrettified("<div><p>a</p></div>", "<div><p>a</p></div>\n")

    def test_nesting_with_tail(self):
        # no attributes
        self.assertPrettified(
            "<root><p></p><!-- #a -->\n\n<p></p></root>",
            "<root><p></p><!-- #a -->\n\n  <p></p></root>\n",
        )
        self.assertPrettified(
            "<root><tal:test />a</root>", "<root><tal:test />a</root>\n"
        )
        self.assertPrettified(
            "<root><tal:test />a </root>", ("<root><tal:test />a", "</root>", "")
        )

        self.assertPrettified(
            "<root><tal:test /> a </root>",
            ("<root><tal:test />", "  a", "</root>", ""),
        )
        self.assertPrettified(
            "<root><tal:test /> a </root>", "<root><tal:test />\n  a\n</root>\n"
        )

    def test_many_children(self):
        """"""
        self.assertPrettified(
            "<root><tal:test /><div></div></root>",
            "<root><tal:test /><div></div></root>\n",
        )
        self.assertPrettified(
            "<root><tal:test /> <div></div></root>",
            "<root><tal:test />\n  <div></div></root>\n",
        )

    def test_boolean_attributes(self):
        """Test attributes without value
        (hidden, required, data-attributes, ...)
        Some of them are rendered valueless, some other not.
        """
        self.assertPrettified(
            '<root data-attribute=""></root>', "<root data-attribute></root>\n"
        )
        self.assertPrettified("<root hidden></root>", "<root hidden></root>\n")
        self.assertPrettified("<root class></root>", '<root class=""></root>\n')

    def test_fix_self_closing(self):
        """Check if open self closing tags are rendered correctly"""
        self.assertPrettified("<input><img><input>", "<input /><img /><input />\n")
        self.assertPrettified("<input><a /><b />", "<input /><a></a><b></b>\n")
        self.assertPrettified("<input><a /><b /></input>", "<input /><a></a><b></b>\n")

    def test_element_repr(self):
        prettifier = ZPrettifier(text="")
        self.assertEqual(repr(prettifier.root), "<pretty:-1:null_tag_name />")

    def test_whitelines_not_stripped(self):
        self.assertPrettified("<root>\n</root>", "<root>\n</root>\n")
        self.assertPrettified(
            "<root>\n    Hello!   \n</root>", "<root>\n    Hello!\n</root>\n"
        )

    def test_text_close_to_an_element(self):
        self.assertPrettified(
            "<root>\n (<a></a>)\n</root>", "<root>\n  (<a></a>)\n</root>\n"
        )

    def test_elements_with_new_lines(self):
        self.assertPrettified("<root\n></root>", "<root></root>\n")
        self.assertPrettified(
            '<root\na="b"\n\n></root>',
            ('<root a="b"></root>\n'),
        )
        self.assertPrettified(
            '<root\na="b"\n\nc="d"></root>',
            ('<root a="b"\n      c="d"\n></root>\n'),
        )

    def test_entities(self):
        self.assertPrettified("<root>&nbsp;</root>", "<root>&nbsp;</root>\n")

    def test_single_quotes_in_attrs(self):
        self.assertPrettified('<root a="\'" />', '<root a="\'"></root>\n')

    def test_ampersand_in_attrs(self):
        self.assertPrettified('<root a="&" />', '<root a="&amp;"></root>\n')
        self.assertPrettified('<root a="foo &" />', '<root a="foo &amp;"></root>\n')
        self.assertPrettified('<root a="& bar" />', '<root a="&amp; bar"></root>\n')
        self.assertPrettified('<root a=";&" />', '<root a=";&amp;"></root>\n')
        self.assertPrettified('<root a="&;" />', '<root a="&amp;;"></root>\n')

    def test_escaped_ampersand_in_attrs(self):
        self.assertPrettified('<root a="&amp;" />', '<root a="&amp;"></root>\n')
        self.assertPrettified('<root a="foo &amp;" />', '<root a="foo &amp;"></root>\n')
        self.assertPrettified('<root a="&amp; bar" />', '<root a="&amp; bar"></root>\n')
        self.assertPrettified('<root a=";&amp;" />', '<root a=";&amp;"></root>\n')
        self.assertPrettified('<root a="&amp;;" />', '<root a="&amp;;"></root>\n')

    def test_ampersand_and_column_in_separate_attrs(self):
        self.assertPrettified(
            '<foo a="&" />\n<tal:bar b=";" />',
            '<foo a="&amp;"></foo>\n<tal:bar b=";" />\n',
        )

    def test_sample_html(self):
        self.prettify("sample_html.html")

    def test_sample_html4(self):
        self.prettify("sample_html4.html")

    def test_sample_html_with_preprocessing_instruction(self):
        self.prettify("sample_html_with_preprocessing_instruction.html")

    def test_sample_pt(self):
        self.prettify("sample_pt.pt")

    def test_text_with_markup(self):
        self.prettify("text_with_markup.md")

    def test_text_file(self):
        self.prettify("sample.txt")



from bs4 import BeautifulSoup
from unittest import TestCase
from zpretty.attributes import PrettyAttributes
from zpretty.elements import PrettyElement


class TestZPrettyAttributess(TestCase):
    """Test zpretty"""

    def get_element(self, text, level=0):
        """Given a text return a PrettyElement"""
        soup = BeautifulSoup(
            "<soup><fake_root>%s</fake_root></soup>" % text, "html.parser"
        )
        return PrettyElement(soup.fake_root.next_element, level)

    def assertPrettifiedAttributes(self, attributes, expected, level=0):
        """Check if the attributes are properly sorted and formatted"""
        if level == 0:
            el = None
        else:
            el = self.get_element("a", level)
        pretty_attribute = PrettyAttributes(attributes, el)
        observed = pretty_attribute()
        self.assertEqual(observed, expected)

    def test_no_attributes(self):
        self.assertPrettifiedAttributes({}, "")
        self.assertPrettifiedAttributes({}, "", level=2)

    def test_one_attribute(self):
        self.assertPrettifiedAttributes({"a": "1"}, 'a="1"')
        self.assertPrettifiedAttributes({"a": "1"}, 'a="1"', level=1)
        self.assertPrettifiedAttributes({"a": "1"}, 'a="1"', level=2)

    def test_value_with_double_quoptes(self):
        self.assertPrettifiedAttributes({"a": '"'}, "a='\"'")

    def test_transform_forbidden_characters(self):
        self.assertPrettifiedAttributes({"a": "> < &"}, 'a="&gt; &lt; &amp;"')

    def test_many_attributes_attribute(self):
        self.assertPrettifiedAttributes({"a": "1", "b": "2"}, 'a="1"\nb="2"')
        self.assertPrettifiedAttributes(
            {"a": "1", "b": "2"}, '    a="1"\n    b="2"', level=1
        )
        self.assertPrettifiedAttributes(
            {"a": "1", "b": "2"}, '      a="1"\n      b="2"', level=2
        )

    def test_tal_define(self):
        self.assertPrettifiedAttributes(
            {"tal:define": "a 1; b 2"},
            "\n".join(('tal:define="', "  a 1;", "  b 2;", '"')),
        )

    def test_format_attributes_many_attribute(self):
        self.assertPrettifiedAttributes(
            {
                "a": "1",
                "b": "2",
                "class": "hidden",
                "tal:define": "a 1; b 2",
                "data-multiline": "foo\n   bar",
            },
            "\n".join(
                (
                    'class="hidden"',
                    'a="1"',
                    'b="2"',
                    'data-multiline="foo',
                    '   bar"',
                    'tal:define="',
                    "  a 1;",
                    "  b 2;",
                    '"',
                )
            ),
        )



from bs4 import BeautifulSoup
from bs4.element import Doctype
from bs4.element import ProcessingInstruction
from logging import getLogger
from uuid import uuid4
from zpretty.elements import PrettyElement

import fileinput
import re


logger = getLogger(__name__)


class ZPrettifier(object):
    """Wraps and renders some text that may contain xml like stuff"""

    pretty_element = PrettyElement
    parser = "html.parser"
    builder = None
    _end_with_newline = True
    _newlines_marker = f"new-line-{str(uuid4())}"
    _ampersand_marker = str(uuid4())
    _cdata_marker = str(uuid4())
    _cdata_pattern = re.compile(r"<!\[CDATA\[(.*?)\]\]>", re.DOTALL)
    _doctype_marker = f"<!DOCTYPE foo-{str(uuid4())}>"
    _doctype_pattern = re.compile(
        r"(<!DOCTYPE[^>[]*(\[[^]]*\])?>)", re.IGNORECASE | re.DOTALL
    )
    _cdatas = []
    _doctype = None

    def __init__(self, filename="", text="", encoding="utf8"):
        """Create a prettifier instance taking the contents
        from a text or a filename
        """
        self._entity_mapping = {}
        self.encoding = encoding
        self.filename = filename
        if self.filename:
            text = "".join(fileinput.input([filename]))
        if not isinstance(text, str):
            text = text.decode(self.encoding)
        self.original_text = text
        self.text = self._prepare_text()
        soup = self.get_soup(self.text)
        # Workaround for https://github.com/collective/zpretty/issues/116
        # restore the ampersands
        # in the attributes so that bogus ones can be escaped
        for el in soup.descendants:
            attrs = getattr(el, "attrs", {})
            for key, value in attrs.items():
                if self._ampersand_marker in value:
                    attrs[key] = value.replace(self._ampersand_marker, "&")

        self.soup = soup

        # Cleanup all spurious self._newlines_marker attributes, see #35
        key = self._newlines_marker.partition("=")[0]
        for el in self.soup.find_all(attrs={key: ""}):
            el.attrs.pop(key, None)

        self.root = self.pretty_element(self.soup, -1)

    def _prepare_text(self):
        """This tweaks the text passed to the prettifier
        to overcome some limitations of the BeautifulSoup parser
        that wants to strip what he does not understand
        (e.g. CDATAs or funny entities).
        """
        text = self.original_text
        self._cdatas = re.findall(self._cdata_pattern, text)
        try:
            self._doctype = self._doctype_pattern.search(text).group(0)
        except AttributeError:
            # No match
            pass
        text = re.sub(self._cdata_pattern, self._cdata_marker, text)
        text = re.sub(self._doctype_pattern, self._doctype_marker, text)

        # Get all the entities in the text and replace them with a marker
        # The text might contain undefined entities that BeautifulSoup
        # will strip out.
        entities = {
            _[0]
            for _ in re.findall(
                r"(&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-fA-F]{1,6});)", text, re.IGNORECASE
            )
        }
        for entity in entities:
            marker = str(uuid4())
            self._entity_mapping[entity] = marker
            text = text.replace(entity, marker)
        return "\n".join(
            line if line.strip() else self._newlines_marker
            for line in text.splitlines()
        ).replace("&", self._ampersand_marker)

    def get_soup(self, text):
        """Tries to get the soup from the given test

        If the text is not some xml like think a dummy element will be used to wrap it.
        """
        original_soup = BeautifulSoup(text, self.parser)
        try:
            first_el = next(original_soup.children)
        except StopIteration:
            first_el = None
        if isinstance(first_el, (Doctype, ProcessingInstruction)):
            return original_soup

        markup = "<{null}>{text}</{null}>".format(
            null=self.pretty_element.null_tag_name, text=text
        )
        wrapped_soup = BeautifulSoup(markup, self.parser)
        return getattr(wrapped_soup, self.pretty_element.null_tag_name)

    def pretty_print(self, el):
        """Pretty print an element indenting it based on level"""
        prettified = (
            el().replace(self._newlines_marker, "").replace(self._ampersand_marker, "&")
        )
        # Restore CDATAs
        for cdata in self._cdatas:
            prettified = prettified.replace(
                self._cdata_marker, f"<![CDATA[{cdata}]]>", 1
            )
        # Restore DocTypes
        if self._doctype:
            prettified = prettified.replace(self._doctype_marker, self._doctype)
        # Restore entities
        for entity, marker in self._entity_mapping.items():
            prettified = prettified.replace(marker, entity)
        if self._end_with_newline and not prettified.endswith("\n"):
            prettified += "\n"
        return prettified

    def check(self):
        """Checks if the input object should be prettified"""
        return self.original_text == self()

    def __call__(self):
        if not self.root.getchildren():
            # The parsed content is not even something that looks like an XML
            return self.original_text
        return self.pretty_print(self.root)



from bs4 import BeautifulSoup
from logging import getLogger
from zpretty.xml import XMLAttributes
from zpretty.xml import XMLElement
from zpretty.xml import XMLPrettifier


logger = getLogger(__name__)


class ZCMLAttributes(XMLAttributes):
    """Customized attribute formatter for zcml"""

    # The ZCML guides wants attributes indented by 4 spaces
    prefix = 4 * " "
    _multiline_attributes = ("for",)
    _xml_attribute_order_by_ns_and_tag = {
        "http://namespaces.plone.org/monkey": {
            # https://github.com/plone/collective.monkeypatcher/blob/0acedea2c0599e3d5a2fbe19df7c2431f34f3fff/src/collective/monkeypatcher/meta.py#L72  # noqa: E501
            "patch": (
                "original",
                "replacement",
                "class",
                "module",
                "handler",
                "preservedoc",
                "docstringWarning",
                "description",
                "order",
                "ignoreOriginal",
                "preserveOriginal",
                "preconditions",
            ),
        },
        "http://namespaces.plone.org/plone": {
            # https://github.com/plone/plone.behavior/blob/3f6d9cbf18e4b3994d02a0875f9fcfc2a1c8d1a6/plone/behavior/metaconfigure.py#L110  # noqa: E501
            "behavior": (
                "name",
                "title",
                "description",
                "factory",
                "provides",
                "for",
                "marker",
                "name_only",
                "former_dotted_names",
            ),
            # https://github.com/plone/plone.rest/blob/ca24e17d972174e5043e27b1bc5da5a752cfb924/src/plone/rest/zcml.py#L200  # noqa: E501
            "CORSPolicy": (
                "allow_origin",
                "allow_credentials",
                "allow_methods",
                "expose_headers",
                "allow_headers",
                "max_age",
                "for",
                "layer",
            ),
            # https://github.com/plone/plone.app.portlets/blob/50ee05b7954a2cfa2d1760e14ea712c9d9dfc791/plone/app/portlets/metaconfigure.py#L27  # noqa: E501
            "portlet": (
                "name",
                "interface",
                "assignment",
                "renderer",
                "addview",
                "editview",
                "view_permission",
                "edit_permission",
            ),
            # https://github.com/plone/plone.app.portlets/blob/50ee05b7954a2cfa2d1760e14ea712c9d9dfc791/plone/app/portlets/metaconfigure.py#L120  # noqa: E501
            "portletRenderer": (
                "portlet",
                "class",
                "template",
                "for",
                "layer",
                "view",
                "manager",
            ),
            # https://github.com/plone/plone.contentrules/blob/0b61d5f7b0b4964167e09f0f3f6454e80c5e1efd/plone/contentrules/rule/metaconfigure.py#L26  # noqa: E501
            "ruleAction": (
                "name",
                "title",
                "description",
                "for",
                "event",
                "schema",
                "factory",
                "addview",
                "editview",
            ),
            # https://github.com/plone/plone.contentrules/blob/0b61d5f7b0b4964167e09f0f3f6454e80c5e1efd/plone/contentrules/rule/metaconfigure.py#L7  # noqa: E501
            "ruleCondition": (
                "name",
                "title",
                "description",
                "for",
                "event",
                "schema",
                "factory",
                "addview",
                "editview",
            ),
            # https://github.com/plone/plone.rest/blob/ca24e17d972174e5043e27b1bc5da5a752cfb924/src/plone/rest/zcml.py#L81  # noqa: E501
            "service": (
                "method",
                "accept",
                "factory",
                "for",
                "permission",
                "layer",
                "name",
            ),
            # https://github.com/plone/plone.resource/blob/e0e3048b778b9d3ef90045f49ea11a168c5cf1e4/plone/resource/zcml.py#L37  # noqa: E501
            "static": ("directory", "name", "type"),
            # https://github.com/plone/plone.tiles/blob/93c1474009f54beb4c713f2054445f49dc1a3f1f/plone/tiles/meta.py#L100  # noqa: E501
            "tile": (
                "name",
                "title",
                "description",
                "for",
                "schema",
                "class",
                "template",
                "permission",
                "add_permission",
                "edit_permission",
                "delete_permission",
                "icon",
                "layer",
            ),
        },
        "http://namespaces.zope.org/browser": {
            # https://github.com/zopefoundation/zope.browsermenu/blob/0409ba2f8575a3997b1fec12fc2e0535625b3b65/src/zope/browsermenu/metaconfigure.py#L247  # noqa: E501
            "addMenuItem": (
                "title",
                "description",
                "menu",
                "factory",
                "for",
                "class",
                "view",
                "icon",
                "filter",
                "permission",
                "layer",
                "extra",
                "order",
                "item_class",
            ),
            # https://github.com/zopefoundation/zope.publisher/blob/af5afbb35e490780c0d02d2f6f08f42ade289ebe/src/zope/publisher/zcml.py#L97  # noqa: E501
            "defaultSkin": ("name",),
            # https://github.com/zopefoundation/zope.publisher/blob/af5afbb35e490780c0d02d2f6f08f42ade289ebe/src/zope/publisher/zcml.py#L106  # noqa: E501
            "defaultView": ("name", "for", "layer"),
            # https://github.com/zopefoundation/zope.browserresource/blob/master/src/zope/browserresource/metaconfigure.py#L124  # noqa: E501
            "icon": (
                "name",
                "title",
                "for",
                "file",
                "resource",
                "layer",
                "width",
                "height",
            ),
            # https://github.com/zopefoundation/zope.browserresource/blob/3b82ea3174aa7a6f241f347cbdfe7eebebcff61d/src/zope/browserresource/metaconfigure.py#L174  # noqa: E501
            "i18n-resource": ("name", "defaultLanguage", "permission", "layer"),
            # https://github.com/zopefoundation/z3c.jbot/blob/c16f9302b96b5994939c877a0fcf28d9739ee921/src/z3c/jbot/metaconfigure.py#L56  # noqa: E501
            "jbot": ("directory", "layer"),
            # https://github.com/zopefoundation/zope.browsermenu/blob/0409ba2f8575a3997b1fec12fc2e0535625b3b65/src/zope/browsermenu/metaconfigure.py  # noqa: E501
            "menu": ("id", "title", "description", "class", "interface"),
            # https://github.com/zopefoundation/zope.browsermenu/blob/0409ba2f8575a3997b1fec12fc2e0535625b3b65/src/zope/browsermenu/metaconfigure.py#L109  # noqa: E501
            "menuItem": (
                "menu",
                "title",
                "description",
                "for",
                "action",
                "icon",
                "filter",
                "permission",
                "layer",
                "extra",
                "order",
                "item_class",
            ),
            # https://github.com/zopefoundation/zope.browsermenu/blob/0409ba2f8575a3997b1fec12fc2e0535625b3b65/src/zope/browsermenu/metaconfigure.py#L151  # noqa: E501
            "menuItems": ("menu", "for", "permission", "layer"),
            # https://github.com/zopefoundation/zope.browserpage/blob/52ca649ab013cdc4e7e306efcd30585da9087272/src/zope/browserpage/metaconfigure.py#L109  # noqa: E501
            "page": (
                "name",
                "for",
                "class",
                "allowed_interface",
                "allowed_attributes",
                "attribute",
                "template",
                "permission",
                "layer",
                "menu",
                "title",
            ),
            # https://github.com/zopefoundation/zope.browserpage/blob/52ca649ab013cdc4e7e306efcd30585da9087272/src/zope/browserpage/metaconfigure.py#L181  # noqa: E501
            "pages": (
                "for",
                "class",
                "allowed_interface",
                "allowed_attributes",
                "permission",
                "layer",
            ),
            # https://github.com/zopefoundation/Zope/blob/e3dbd32848e3ea97e2146a9fa690c63555bf61a5/src/Products/Five/fiveconfigure.py#L32  # noqa: E 501
            "pagesFromDirectory": ("directory", "module", "for", "permission", "layer"),
            # https://github.com/zopefoundation/zope.browserresource/blob/3b82ea3174aa7a6f241f347cbdfe7eebebcff61d/src/zope/browserresource/metaconfigure.py#L54  # noqa: E501
            "resource": (
                "name",
                "factory",
                "file",
                "image",
                "template",
                "permission",
                "layer",
            ),
            # https://github.com/zopefoundation/zope.browserresource/blob/3b82ea3174aa7a6f241f347cbdfe7eebebcff61d/src/zope/browserresource/metaconfigure.py#L102  # noqa: E501
            "resourceDirectory": ("name", "directory", "layer", "permission"),
            # https://github.com/zopefoundation/Zope/blob/e3dbd32848e3ea97e2146a9fa690c63555bf61a5/src/Products/Five/sizeconfigure.py  # noqa: E501
            "sizable": ("class",),
            # https://github.com/zopefoundation/z3c.jbot/blob/c16f9302b96b5994939c877a0fcf28d9739ee921/src/z3c/jbot/metaconfigure.py#L56  # noqa: E501
            "templateOverrides": ("directory", "layer"),
            # https://github.com/zopefoundation/zope.browserresource/blob/3b82ea3174aa7a6f241f347cbdfe7eebebcff61d/src/zope/browserresource/metaconfigure.py#L188  # noqa: E501
            "translation": ("language", "file", "image"),
            # https://github.com/zopefoundation/zope.browserpage/blob/52ca649ab013cdc4e7e306efcd30585da9087272/src/zope/browserpage/metaconfigure.py#L210  # noqa: E501
            "view": (
                "name",
                "provides",
                "for",
                "class",
                "allowed_interface",
                "allowed_attributes",
                "permission",
                "layer",
                "menu",
                "title",
            ),
            # https://github.com/zopefoundation/zope.viewlet/blob/d57ed30b35ca3f0d821cbb673513e63ee0747ffa/src/zope/viewlet/metaconfigure.py#L92  # noqa: E501
            "viewlet": (
                "name",
                "for",
                "view",
                "manager",
                "class",
                "allowed_interface",
                "allowed_attributes",
                "attribute",
                "template",
                "permission",
                "layer",
            ),
            # https://github.com/zopefoundation/zope.viewlet/blob/d57ed30b35ca3f0d821cbb673513e63ee0747ffa/src/zope/viewlet/metaconfigure.py#L28  # noqa: E501
            "viewletManager": (
                "name",
                "provides",
                "for",
                "view",
                "class",
                "allowed_interface",
                "allowed_attributes",
                "template",
                "permission",
                "layer",
            ),
        },
        "http://namespaces.zope.org/cache": {
            # https://github.com/zopefoundation/z3c.caching/blob/bd1a6cac1bef90b2d5c3c7cb783ae335feee4dff/src/z3c/caching/zcml.py#L36  # noqa: E501
            "ruleset": ("for", "ruleset"),
            # https://github.com/zopefoundation/z3c.caching/blob/bd1a6cac1bef90b2d5c3c7cb783ae335feee4dff/src/z3c/caching/zcml.py#L23  # noqa: E501
            "rulesetType": ("name", "title", "description"),
        },
        "http://namespaces.zope.org/cmf": {
            # https://github.com/zopefoundation/Products.CMFCore/blob/147efe3192b659211d07997e1923bd7de0c32361/Products/CMFCore/zcml.py#L64  # noqa: E501
            "registerDirectory": ("name", "directory", "recursive", "ignore"),
        },
        "http://namespaces.zope.org/five": {
            # TODO: OFS/meta.zcml
        },
        "http://namespaces.zope.org/ftw.upgrade": {
            # https://github.com/4teamwork/ftw.upgrade/blob/c003c2f5c2ef63eef9efd23cdf754174a66ebabe/ftw/upgrade/directory/zcml.py#L38  # noqa: E501
            "directory": ("profile", "directory", "soft_dependencies"),
            # https://github.com/4teamwork/ftw.upgrade/blob/c003c2f5c2ef63eef9efd23cdf754174a66ebabe/ftw/upgrade/zcml.py#L52  # noqa: E501
            "importProfile": (
                "title",
                "description",
                "profile",
                "source",
                "destination",
                "directory",
                "handler",
            ),
        },
        "http://namespaces.zope.org/genericsetup": {
            # https://github.com/zopefoundation/Products.GenericSetup/blob/c46e0a4f026fc1355f8a3c002f1515d4c8f63715/Products/GenericSetup/zcml.py#L136  # noqa: E501
            "exportStep": ("name", "title", "description", "handler"),
            # https://github.com/zopefoundation/Products.GenericSetup/blob/c46e0a4f026fc1355f8a3c002f1515d4c8f63715/Products/GenericSetup/zcml.py#L183  # noqa: E501
            "importStep": ("name", "title", "description", "handler"),
            # https://github.com/zopefoundation/Products.GenericSetup/blob/c46e0a4f026fc1355f8a3c002f1515d4c8f63715/Products/GenericSetup/zcml.py#L89  # noqa: E501
            "registerProfile": (
                "name",
                "title",
                "description",
                "provides",
                "for",
                "directory",
                "pre_handler",
                "post_handler",
            ),
            # https://github.com/zopefoundation/Products.GenericSetup/blob/c46e0a4f026fc1355f8a3c002f1515d4c8f63715/Products/GenericSetup/zcml.py#L322  # noqa: E501
            "upgradeDepends": (
                "title",
                "description",
                "profile",
                "source",
                "destination",
                "import_profile",
                "import_steps",
                "run_deps",
                "purge",
                "checker",
                "sortkey",
            ),
            # https://github.com/zopefoundation/Products.GenericSetup/blob/c46e0a4f026fc1355f8a3c002f1515d4c8f63715/Products/GenericSetup/zcml.py#L310  # noqa: E501
            "upgradeStep": (
                "title",
                "description",
                "profile",
                "source",
                "destination",
                "handler",
                "checker",
                "sortkey",
            ),
            # https://github.com/zopefoundation/Products.GenericSetup/blob/c46e0a4f026fc1355f8a3c002f1515d4c8f63715/Products/GenericSetup/zcml.py#L343  # noqa: E501
            "upgradeSteps": ("profile", "source", "destination", "sortkey"),
        },
        "http://namespaces.zope.org/i18n": {
            # https://github.com/zopefoundation/zope.i18n/blob/59d212f62d20f9bad168a0dcb8bb137bd625130b/src/zope/i18n/zcml.py#L78  # noqa: E501
            "registerTranslations": ("directory", "domain"),
        },
        "http://namespaces.zope.org/pluggableauthservice": {
            # https://github.com/zopefoundation/Products.PluggableAuthService/blob/f1d7aac8d76ba4840d46fc1e144083bf103d9c9a/Products/PluggableAuthService/zcml.py#L44  # noqa: E501
            "registerMultiPlugin": ("class", "meta_type"),
        },
        "http://namespaces.zope.org/z3c": {
            # https://github.com/zopefoundation/z3c.form/blob/440eb18c541b88f64e88be9b66fbc27748f75af2/src/z3c/form/zcml.py#L133  # noqa: E501
            "objectWidgetTemplate": (
                "for",
                "view",
                "schema",
                "field",
                "widget",
                "template",
                "layer",
            ),
            # https://github.com/zopefoundation/z3c.form/blob/440eb18c541b88f64e88be9b66fbc27748f75af2/src/z3c/form/zcml.py#L114  # noqa: E501
            "widgetLayout": (
                "for",
                "view",
                "field",
                "widget",
                "template",
                "layer",
                "mode",
                "contentType",
            ),
            # https://github.com/zopefoundation/z3c.form/blob/440eb18c541b88f64e88be9b66fbc27748f75af2/src/z3c/form/zcml.py#L96  # noqa: E501
            "widgetTemplate": (
                "for",
                "view",
                "field",
                "widget",
                "template",
                "layer",
                "mode",
                "contentType",
            ),
        },
        "http://namespaces.zope.org/zope": {
            # https://github.com/zopefoundation/zope.component/blob/e7fe69a371afd52781a3c49c130768b607bcff48/src/zope/component/zcml.py#L164  # noqa: E501
            "adapter": (
                "factory",
                "provides",
                "for",
                "name",
                "permission",
                "trusted",
                "locate",
            ),
            # https://github.com/zopefoundation/zope.configuration/blob/022ff7a08af430f9159ffb1574fd908c17745b45/src/zope/configuration/xmlconfig.py#L561  # noqa: E501
            "exclude": ("package", "package", "files"),
            # https://github.com/zopefoundation/zope.configuration/blob/022ff7a08af430f9159ffb1574fd908c17745b45/src/zope/configuration/xmlconfig.py#L524  # noqa: E501
            "include": ("package", "file", "files"),
            # https://github.com/zopefoundation/zope.configuration/blob/022ff7a08af430f9159ffb1574fd908c17745b45/src/zope/configuration/xmlconfig.py#L594  # noqa: E501
            "includeOverrides": ("package", "file", "files"),
            # https://github.com/zopefoundation/zope.component/blob/e7fe69a371afd52781a3c49c130768b607bcff48/src/zope/component/zcml.py#L429  # noqa: E501
            "interface": ("interface", "type", "name"),
            # https://github.com/zopefoundation/AccessControl/blob/458139e987ee79c51438ddb7342300a673222285/src/AccessControl/security.py#L177  # noqa: E501
            "permission": ("id", "title", "description"),
            # https://github.com/zopefoundation/zope.component/blob/e7fe69a371afd52781a3c49c130768b607bcff48/src/zope/component/zcml.py#L616  # noqa: E501
            "resource": (
                "factory",
                "provides",
                "type",
                "name",
                "permission",
                "allowed_interface",
                "allowed_attributes",
            ),
            # https://github.com/zopefoundation/AccessControl/blob/458139e987ee79c51438ddb7342300a673222285/src/AccessControl/security.py#L202  # noqa: E501
            "role": ("name",),
            # https://github.com/zopefoundation/zope.component/blob/e7fe69a371afd52781a3c49c130768b607bcff48/src/zope/component/zcml.py#L295  # noqa: E501
            "subscriber": (
                "factory",
                "provides",
                "for",
                "handler",
                "permission",
                "trusted",
                "locate",
            ),
            # https://github.com/zopefoundation/zope.component/blob/e7fe69a371afd52781a3c49c130768b607bcff48/src/zope/component/zcml.py#L373 # noqa: E501
            "utility": (
                "factory",
                "provides",
                "name",
                "component",
                "permission",
            ),
            # https://github.com/zopefoundation/zope.component/blob/e7fe69a371afd52781a3c49c130768b607bcff48/src/zope/component/zcml.py#L523 # noqa: E501
            "view": (
                "name",
                "factory",
                "allowed_interface",
                "allowed_attributes",
                "provides",
                "for",
                "permission",
                "type",
            ),
        },
    }

    # This will be used when no suitable match is found
    _xml_attribute_order_fallback = (
        "name",
        "title",
        "description",
        "package",
        "file",
        "for",
        "provides",
        "factory",
        "manager",
        "class",
        "allowed_attributes",
        "attribute",
        "template",
        "permission",
        "layer",
    )

    @property
    def _xml_attribute_order(self):
        """Sort the attributes based on the element

        _xml_attribute_order_by_ns_and_tag comments contain references
        to the directives declaration
        The attribute order is inspired by Python code signature of the directives,
        but given it is highly inconsistent some arbitrary decisions have been made.

        In particular:
        - optional element (like layer) are sorted after the others
        - class related attributes in views (i.e. allowed_interfaces)
          are sorted right after the class attribute
        - name, title, description are usually sorted first (except for view
          where the title is sorted after the menu attribute)
        """
        try:
            ns = self.element.context.namespace
            name = self.element.context.name
        except AttributeError:
            ns = ""
            name = ""

        mapping_by_namespace = self._xml_attribute_order_by_ns_and_tag.get(ns, {})
        return mapping_by_namespace.get(name, self._xml_attribute_order_fallback)

    def format_multiline(self, name, value):
        """We have two cases according if we have just one attribute or more

        1. single attribute
          1.1 the the element prefix (if any)
          1.2 as many spaces as the element name
          1.2 as many spaces as the attribute name
          1.3 four spaces spaces (leading `<` and space between element and attribute
              the `="`)

        2. many attributes we need to add:
          2.1 the the element prefix (if any)
          2.2 attribute indentation (4 spaces)
          2.3 as many spaces as the attribute name
          2.4 2 spaces to take into account the leading `="`
        """
        value_lines = filter(None, value.split())
        if len(self) == 1:
            line_joiner = (
                "\n"
                + getattr(self.element, "prefix", "")
                + " " * len(getattr(self.element, "tag", ""))
                + " " * len(name)
                + " " * 4
            )
        else:
            line_joiner = (
                "\n"
                + getattr(self.element, "prefix", "")
                + self.prefix
                + " " * len(name)
                + " " * 2
            )
        return line_joiner.join(value_lines)

    def lstrip(self):
        """Actually we do not want to remove the spaces"""
        return self()

    def __call__(self):
        """Render the attributes as text

        Render and an empty string if no attributes
        If we have one attribute we do not indent it
        If we have many we indent them by 4 spaces + the indentation of the element
        """
        if len(self) == 0:
            return ""
        if len(self) == 1:
            for line in self.lines():
                return line

        if self.element:
            prefix = self.element.prefix + self.prefix
        else:
            prefix = self.prefix

        lines = self.lines()
        return prefix + f"\n{prefix}".join(lines)


class ZCMLElement(XMLElement):
    before_closing_multiline = "    "
    attribute_klass = ZCMLAttributes

    self_closing_multiline_template = "\n".join(
        ("{prefix}<{tag}\n{attributes}", "{prefix}{before_closing_multiline}/>")
    )
    start_tag_multiline_template = "\n".join(
        ("{prefix}<{tag}\n{attributes}", "{prefix}{before_closing_multiline}>")
    )


class ZCMLPrettifier(XMLPrettifier):
    """Prettify according to the ZCML style guide"""

    pretty_element = ZCMLElement

    def get_soup(self, text):
        """Tries to get the soup from the given text"""
        markup = "<{null}>{text}</{null}>".format(
            null=self.pretty_element.null_tag_name, text=text
        )
        wrapped_soup = BeautifulSoup(markup, self.parser)
        return getattr(wrapped_soup, self.pretty_element.null_tag_name)



from setuptools import find_packages
from setuptools import setup

import os


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


setup(
    name="zpretty",
    version="3.1.0a2",
    description="An opinionated HTML/XML soup formatter",
    keywords=["Formatter", "HTML", "Prettifier", "Pretty print", "TAL", "XML", "ZPT"],
    long_description="\n".join((read("README.md"), read("HISTORY.md"))),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
    author="Alessandro Pisa",
    author_email="alessandro.pisa@gmail.com",
    url="https://github.com/collective/zpretty",
    license="BSD",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=["setuptools", "beautifulsoup4", "lxml"],
    extras_require={
        "test": ["pre-commit", "pytest-cov", "pytest"],
        "development": ["zest.releaser", "check-manifest", "pyroma"],
    },
    test_suite="zpretty.tests",
    entry_points={"console_scripts": ["zpretty=zpretty.cli:run"]},
    include_package_data=True,
    zip_safe=False,
)



from argparse import ArgumentParser
from os.path import splitext
from pathlib import Path
from sys import stderr
from sys import stdout
from zpretty.prettifier import ZPrettifier
from zpretty.xml import XMLPrettifier
from zpretty.zcml import ZCMLPrettifier

import re


try:
    # Python >= 3.8
    from importlib.metadata import version

    version = version("zpretty")
except ImportError:
    # Python < 3.8
    from pkg_resources import get_distribution

    version = get_distribution("zpretty").version


class CLIRunner:
    """A class to run zpretty from the command line"""

    _default_include = r"\.(html|pt|xml|zcml)$"
    _default_exclude = (
        r"/(\.direnv|\.eggs|\.git|\.hg|\.mypy_cache|\.nox|\.tox|\.venv|venv|"
        r"\.svn|\.ipynb_checkpoints|_build|buck-out|build|dist|__pypackages__)/"
    )

    def __init__(self):
        self.errors = []
        self.config = self.parser.parse_args()

    @property
    def parser(self):
        """The parser we are using to parse the command line arguments"""
        parser = ArgumentParser(
            prog="zpretty",
            description="An opinionated HTML/XML soup formatter",
            epilog=f"The default exclude pattern is: `{self._default_exclude}`",
        )
        parser.add_argument(
            "--encoding",
            help="The file encoding (defaults to utf8)",
            action="store",
            dest="encoding",
            default="utf8",
        )
        parser.add_argument(
            "-i",
            "--inplace",
            help="Format files in place (overwrite existing file)",
            action="store_true",
            dest="inplace",
            default=False,
        )
        parser.add_argument(
            "-v",
            "--version",
            help="Show zpretty version number",
            action="version",
            version=f"zpretty {version}",
        )
        parser.add_argument(
            "-x",
            "--xml",
            help="Treat the input file(s) as XML",
            action="store_true",
            dest="xml",
            default=False,
        )
        parser.add_argument(
            "-z",
            "--zcml",
            help="Treat the input file(s) as XML. Follow the ZCML styleguide",
            action="store_true",
            dest="zcml",
            default=False,
        )
        parser.add_argument(
            "--check",
            help=(
                "Return code 0 if nothing would be changed, "
                "1 if some files would be reformatted"
            ),
            action="store_true",
            dest="check",
            default=False,
        )
        parser.add_argument(
            "--include",
            help=(
                f"A regular expression that matches files and "
                f" directories that should be included on recursive searches. "
                f"An empty value means all files are included regardless of the name. "
                f"Use forward slashes for directories on all platforms (Windows, too). "
                f"Exclusions are calculated first, inclusions later. "
                f"[default: {self._default_include}]"
            ),
            action="store",
            dest="include",
            default=self._default_include,
        )
        parser.add_argument(
            "--exclude",
            help=(
                f"A regular expression that matches files and "
                f"directories that should be excluded on "
                f"recursive searches. An empty value means no "
                f"paths are excluded. Use forward slashes for "
                f"directories on all platforms (Windows, too). "
                f"Exclusions are calculated first, inclusions "
                f"later. [default: {self._default_exclude}] "
            ),
            action="store",
            dest="exclude",
            default=self._default_exclude,
        )

        parser.add_argument(
            "--extend-exclude",
            help=(
                "Like --exclude,  but adds additional files "
                "and directories on top of the excluded ones. "
                "(Useful if you simply want to add to the default)"
            ),
            action="store",
            dest="extend_exclude",
            default=None,
        )
        parser.add_argument(
            "paths",
            nargs="*",
            default="-",
            help="The list of files or directory to prettify (defaults to stdin). "
            "If a directory is passed, all files and directories matching the regular "
            "expression passed to --include will be prettified.",
        )
        return parser

    def choose_prettifier(self, path):
        """Choose the best prettifier given the config and the input file"""
        config = self.config
        if config.zcml:
            return ZCMLPrettifier
        if config.xml:
            return XMLPrettifier
        ext = splitext(path)[-1].lower()
        if ext == ".xml":
            return XMLPrettifier
        if ext == ".zcml":
            return ZCMLPrettifier
        return ZPrettifier

    @property
    def good_paths(self):
        """Return a list of good paths"""
        good_paths = []

        try:
            exclude = re.compile(self.config.exclude)
        except re.error:
            exclude = re.compile(self._default_exclude)
            self.errors.append(
                f"Invalid regular expression for --exclude: {self.config.exclude!r}"
            )

        try:
            extend_exclude = self.config.extend_exclude and re.compile(
                self.config.extend_exclude
            )
        except re.error:
            extend_exclude = None
            self.errors.append(
                f"Invalid regular expression for --extend-exclude: "
                f"{self.config.extend_exclude!r}"
            )

        try:
            include = re.compile(self.config.include)
        except re.error:
            include = re.compile(self._default_include)
            self.errors.append(
                f"Invalid regular expression for --include: {self.config.include!r}"
            )

        for path in self.config.paths:
            # use Pathlib to check if the file exists and it is a file
            if path == "-":
                good_paths.append(path)
                continue
            if exclude.match(path) or (extend_exclude and extend_exclude.match(path)):
                continue

            path_instance = Path(path)
            if path_instance.is_file():
                good_paths.append(path)
            elif path_instance.is_dir():
                for file in path_instance.glob("**/*"):
                    if file.is_file():
                        if (
                            include.search(str(file))
                            and not exclude.search(str(file))
                            and not (
                                extend_exclude and extend_exclude.search(str(file))
                            )
                        ):
                            good_paths.append(str(file))
            else:
                self.errors.append(f"Cannot open: {path}")

        return sorted(good_paths)

    def run(self):
        """Prettify each filename passed in the command line"""
        encoding = self.config.encoding
        for path in self.good_paths:
            # use Pathlib to check if the file exists and it is a file
            Prettifier = self.choose_prettifier(path)
            prettifier = Prettifier(path, encoding=encoding)
            if self.config.check:
                if not prettifier.check():
                    self.errors.append(f"This file would be rewritten: {path}")
                continue
            prettified = prettifier()
            if self.config.inplace and not path == "-":
                with open(path, "w") as f:
                    f.write(prettified)
                continue
            stdout.write(prettified)

        if self.errors:
            message = "\n".join(self.errors)
            stderr.write(f"{message}\n")
            exit(1)


def run():
    CLIRunner().run()  # pragma: no cover


if __name__ == "__main__":
    run()  # pragma: no cover
